#include "charsets.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <zarchiver.h>

typedef struct {
  unsigned short indx; /* index into big table */
  unsigned short used; /* bitmask of used entries */
} Summary16;

#include "codec/ascii.h"
#include "codec/jisx0201.h"
#include "codec/jisx0208.h"
#include "codec/gb2312.h"
#include "codec/gbk.h"
#include "codec/ces_gbk.h"
#include "codec/big5.h"
#include "codec/ksc5601.h"
#include "codec/euc_kr.h"

//775 853 855 856 861 874 922 943
#include "codec/cp866.h" //Cyrillic
#include "codec/cp863.h" //French Canada
#include "codec/cp857.h" //Turkish
#include "codec/cp932.h" //Japanese, Shift-JIS
#include "codec/cp936.h" //Simplified Chinese
#include "codec/cp949.h" //Korean (Unified Hangeul Code)
#include "codec/cp950.h" //Traditional Chinese
#include "codec/cp850.h"
#include "codec/cp852.h"
#include "codec/cp858.h"
#include "codec/cp860.h"//Portuguese
#include "codec/cp862.h"//Hebrew
#include "codec/cp865.h"//Nordic
#include "codec/cp737.h"//Greek (formerly 437G)
#include "codec/cp869.h"//Modern Greek
#include "codec/cp874.h"//Thai
#include "codec/cp775.h"
#include "codec/cp855.h"
#include "codec/cp720.h"//arabic dos

#include "codec/cp1125.h"// (OEM Ukraine)
#include "codec/cp1250.h"// (ANSI Central European)
#include "codec/cp1251.h"// (ANSI Cyrillic)
#include "codec/cp1252.h"// (ANSI Latin 1, Western European)
#include "codec/cp1253.h"// (ANSI Greek)
#include "codec/cp1254.h"// (ANSI Turkish)
#include "codec/cp1255.h"// (ANSI Hebrew)
#include "codec/cp1256.h"// (ANSI Arabic)
#include "codec/cp1257.h"// (ANSI Baltic)
#include "codec/cp1258.h"// (ANSI/OEM Vietnamese)
#include "codec/cp437.h" //US

#include "codec/utf8.h"

static const _mbtowc codec_m[] = {
	cp866_mbtowc,cp720_mbtowc,cp863_mbtowc,cp857_mbtowc,cp932_mbtowc,
	cp936_mbtowc,cp949_mbtowc,cp950_mbtowc,cp850_mbtowc,cp852_mbtowc,
	cp858_mbtowc,cp860_mbtowc,cp862_mbtowc,cp865_mbtowc,cp737_mbtowc,
	cp869_mbtowc,cp775_mbtowc,cp855_mbtowc,cp874_mbtowc,
	cp1125_mbtowc,cp1250_mbtowc,cp1251_mbtowc,cp1252_mbtowc,
	cp1253_mbtowc,cp1254_mbtowc,cp1255_mbtowc,cp1256_mbtowc,
	cp1257_mbtowc,cp1258_mbtowc,cp437_mbtowc, 
	
	utf8_mbtowc, 
	euc_kr_mbtowc,
	};
static const _wctomb codec_w[] = {
	cp866_wctomb,cp720_wctomb,cp863_wctomb,cp857_wctomb,cp932_wctomb,
	cp936_wctomb,cp949_wctomb,cp950_wctomb,cp850_wctomb,cp852_wctomb,
	cp858_wctomb,cp860_wctomb,cp862_wctomb,cp865_wctomb,cp737_wctomb,
	cp869_wctomb,cp775_wctomb,cp855_wctomb,cp874_wctomb,
	cp1125_wctomb,cp1250_wctomb,cp1251_wctomb,cp1252_wctomb,
	cp1253_wctomb,cp1254_wctomb,cp1255_wctomb,cp1256_wctomb,
	cp1257_wctomb,cp1258_wctomb,cp437_wctomb, 
	
	utf8_wctomb,
	NULL,
	};

static const uint16_t codec_n[] = {
	866,720,863,857,932,
	936,949,950,850,852,
	858,860,862,865,737, // -858
	869,775,855,874, // -869
	1125,1250,1251,1252,
	1253,1254,1255,1256,
	1257,1258,437,
	
	//UTF codepage:
	CODEPAGE_UTF(8),
	CODEPAGE_EUCKR,
	};

_mbtowc __UI get_mbtowc(int cp, int def)
{
	int i;
	
	for (i=0;i<(sizeof(codec_n)/sizeof(uint16_t));i++)
		if (codec_n[i] == cp)
			return codec_m[i];
	
	for (i=0;i<(sizeof(codec_n)/sizeof(uint16_t));i++)
		if (codec_n[i] == def)
			return codec_m[i];
	
	return NULL;
}

_wctomb __UI get_wctomb(int cp, int def)
{
	int i;
	
	for (i=0;i<(sizeof(codec_n)/sizeof(uint16_t));i++)
		if (codec_n[i] == cp)
			return codec_w[i];

	for (i=0;i<(sizeof(codec_n)/sizeof(uint16_t));i++)
		if (codec_n[i] == def)
			return codec_w[i];

	return NULL;
}

size_t __UI iconv(int from, int to, const char **inbuf, size_t *inbytesleft, char **outbuf, size_t *outbytesleft)
{
	if (outbuf == NULL || outbytesleft == NULL)
		return (size_t)(int)-1;
	
	_mbtowc mbtowc =  get_mbtowc(from, 0);
	_wctomb wctomb = get_wctomb(to, 0);
	if (mbtowc == NULL || wctomb == NULL)
		return (size_t)(int)-1;
	
	const char *in = *inbuf;
	size_t len;
	if (inbytesleft)
		len = *inbytesleft;
	else
		len = strlen(in);
	
	char *out = *outbuf;
	size_t out_left = *outbytesleft;
	
	wchar_t tmp, lastM, lastW;
	size_t i;
	int result;
	i=0;
	lastM = lastW = 0;
	while (i < len && out_left>0)
	{
		result = mbtowc(&lastM, &tmp, (const unsigned char *)in, len-i);
		if (result<0)
			break;
		i+=result;
		in+=result;
		
		result = wctomb(&lastW, (unsigned char *)out, tmp, out_left);
		if (result<0)
			break;
		out_left-=result;
		out+=result;
	}
	*out = '\0';
	
	if (result < 0)
		return (size_t)(int)-1;
	
	*outbytesleft -= out_left;
	return *outbytesleft;
}
